from typing import Set, Tuple, Optional

from cbrrr import CID

from .node import MSTNode
from .node_store import NodeStore
from .node_walker import NodeWalker

class InvalidProof(Exception):
	pass

# works for both inclusion and exclusion proofs
def find_rpath_and_build_proof(ns: NodeStore, root_cid: CID, rpath: str) -> Tuple[Optional[CID], Set[CID]]:
	walker = NodeWalker(ns, root_cid)
	value = walker.find_rpath(rpath) # returns None if not found
	proof = {frame.node.cid for frame in walker.stack}
	return value, proof

def verify_inclusion(ns: NodeStore, root_cid: CID, rpath: str) -> None:
	walker = NodeWalker(ns, root_cid)
	try:
		if walker.find_rpath(rpath) is None:
			raise InvalidProof("rpath not present in MST")
	except KeyError:
		raise InvalidProof("missing MST blocks")

def verify_exclusion(ns: NodeStore, root_cid: CID, rpath: str) -> None:
	walker = NodeWalker(ns, root_cid)
	try:
		if walker.find_rpath(rpath) is not None:
			raise InvalidProof("rpath *is* present in MST")
	except KeyError:
		raise InvalidProof("missing MST blocks")



import unittest
import io

from atmst.blockstore.car_file import decode_varint, encode_varint

class MSTDiffTestCase(unittest.TestCase):
	def test_varint_encode(self):
		self.assertEqual(encode_varint(0), b"\x00")
		self.assertEqual(encode_varint(1), b"\x01")
		self.assertEqual(encode_varint(127), b"\x7f")
		self.assertEqual(encode_varint(128), b"\x80\x01")
		self.assertEqual(encode_varint(2**63-1), b'\xff\xff\xff\xff\xff\xff\xff\xff\x7f')
		self.assertRaises(ValueError, encode_varint, 2**63)
		self.assertRaises(ValueError, encode_varint, -1)

	def test_varint_decode(self):
		self.assertEqual(decode_varint(io.BytesIO(b"\x00")), 0)
		self.assertEqual(decode_varint(io.BytesIO(b"\x01")), 1)
		self.assertEqual(decode_varint(io.BytesIO(b"\x7f")), 127)
		self.assertEqual(decode_varint(io.BytesIO(b"\x80\x01")), 128)
		self.assertEqual(decode_varint(io.BytesIO(b'\xff\xff\xff\xff\xff\xff\xff\xff\x7f')), 2**63-1)
		self.assertRaises(ValueError, decode_varint, io.BytesIO(b'\xff\xff\xff\xff\xff\xff\xff\xff\xff\x7f')) # too big
		self.assertRaises(ValueError, decode_varint, io.BytesIO(b"")) # too short
		self.assertRaises(ValueError, decode_varint, io.BytesIO(b'\xff')) # truncated
		self.assertRaises(ValueError, decode_varint, io.BytesIO(b"\x80\x00")) # not minimally encoded

if __name__ == '__main__':
	unittest.main(module="tests.test_varint")



import unittest
import random
import string

from atmst.all import MemoryBlockStore, NodeStore, NodeWrangler, mst_diff, very_slow_mst_diff
from atmst.mst.node import MSTNode
from cbrrr import CID


class MSTDiffTestCase(unittest.TestCase):
	def setUp(self):
		dummy_value = CID.cidv1_dag_cbor_sha256_32_from(b"value")

		bs = MemoryBlockStore()
		self.ns = NodeStore(bs)
		wrangler = NodeWrangler(self.ns)
		root = self.ns.get_node(None).cid

		random.seed(4) # chosen by fair dice roll
		self.included_rpaths = []
		for _ in range(1000):
			random_rpath = "".join([random.choice(string.ascii_lowercase + string.digits) for _ in range(10)])
			self.included_rpaths.append(random_rpath)
			root = wrangler.put_record(root, random_rpath, dummy_value)

	def test_inclusion_proof(self):
		pass

if __name__ == "__main__":
	unittest.main(module="tests.test_mst_proofs")



import sys
import os
import base64
import json
from typing import Tuple, BinaryIO

from cbrrr import encode_dag_cbor, decode_dag_cbor, CID

from .blockstore.car_file import ReadOnlyCARBlockStore, encode_varint
from .blockstore import OverlayBlockStore
from .mst.node_store import NodeStore
from .mst.node_walker import NodeWalker
from .mst.diff import mst_diff, record_diff


def prettify_record(record) -> str:
	return json.dumps(record, indent="  ")

def open_car(car_path: str) -> Tuple[ReadOnlyCARBlockStore, dict]:
	carfile = open(car_path, "rb")
	bs = ReadOnlyCARBlockStore(carfile)
	commit = decode_dag_cbor(bs.get_block(bytes(bs.car_root)))
	return bs, commit

def print_info(car_path: str) -> None:
	print(f"Reading {car_path!r}")
	print(f"Size on disk: {os.stat(car_path).st_size} bytes")
	bs, commit = open_car(car_path)
	print("Total CAR blocks:", len(bs.block_offsets))
	print("Root CID:", bs.car_root.encode("base32"))
	print()
	print("ATProto commit info:")
	print("Version:", commit["version"])
	if commit["version"] != 3:
		print(f"Error: only v3 repo format is supported. Got:", commit["version"])
		return
	print("Repo:", commit["did"])
	print("Rev:", commit["rev"])
	print("Sig:", base64.urlsafe_b64encode(commit["sig"]).decode())
	print("MST root:", commit["data"].encode("base32"))

def print_all_records(car_path: str, to_json: bool) -> None:
	bs, commit = open_car(car_path)
	for k, v in NodeWalker(NodeStore(bs), commit["data"]).iter_kv():
		if to_json:
			record = decode_dag_cbor(bs.get_block(bytes(v)), atjson_mode=True)
			print(f"{json.dumps(k)} -> {prettify_record(record)}")
		else:
			print(f"{json.dumps(k)} -> {v.encode('base32')}")

def list_all(car_path: str):
	print_all_records(car_path, to_json=False)

def dump_all(car_path: str):
	sys.setrecursionlimit(99999999) # allow printing very deeply nested records
	print_all_records(car_path, to_json=True)

def dump_record(car_path: str, key: str):
	bs, commit = open_car(car_path)
	val = NodeWalker(NodeStore(bs), commit["data"]).find_rpath(key)
	if val is None:
		print("Record not found!", file=sys.stderr)
		sys.exit(-1)
	record = decode_dag_cbor(bs.get_block(bytes(val)), atjson_mode=True)
	print(prettify_record(record))

def write_block(file: BinaryIO, data: bytes) -> None:
	file.write(encode_varint(len(data)))
	file.write(data)

def compact(car_in: str, car_out: str):
	bs, commit = open_car(car_in)
	with open(car_out, "wb") as carfile_out:
		new_header = encode_dag_cbor({
			"version": 1,
			"roots": [bs.car_root]
		})
		write_block(carfile_out, new_header)
		write_block(carfile_out, bytes(bs.car_root) + encode_dag_cbor(commit))
		dedup = {bs.car_root}

		for node in NodeWalker(NodeStore(bs), commit["data"]).iter_nodes():
			if node.cid not in dedup:
				write_block(carfile_out, bytes(node.cid) + node.serialised)
				dedup.add(node.cid)
			for v in node.vals:
				if v not in dedup:
					write_block(carfile_out, bytes(v) + bs.get_block(bytes(v)))
					dedup.add(v)

def _delta_str(a: str, b: str):
	if a == b:
		return f"{a} == {b}"
	return f"{a} -> {b}"

def print_record_diff(car_a: str, car_b: str):
	bs_a, commit_a = open_car(car_a)
	bs_b, commit_b = open_car(car_b)
	print(f"Repo: {_delta_str(commit_a['did'], commit_b['did'])}")
	print(f"Revision: {_delta_str(commit_a['rev'], commit_b['rev'])}")
	print(f"Commit: {_delta_str(bs_a.car_root.encode('base32'), bs_b.car_root.encode('base32'))}")
	print(f"MST root: {_delta_str(commit_a['data'].encode('base32'), commit_b['data'].encode('base32'))}")
	print("")
	print("Record delta:")
	bs = OverlayBlockStore(bs_a, bs_b)
	ns = NodeStore(bs)
	mst_created, mst_deleted = mst_diff(ns, commit_a["data"], commit_b["data"])
	for delta in record_diff(ns, mst_created, mst_deleted):
		print(delta)

COMMANDS = {
	"info": (print_info, "print CAR header and repo info"),
	"list": (list_all, "list all records in the CAR (values as CIDs)"),
	"dump": (dump_all, "dump all records in the CAR (values as JSON)"),
	"dump_record": (dump_record, "dump a single record, keyed on ('collection/rkey')"),
	"compact": (compact, "rewrite the whole CAR, dropping any duplicated or unreferenced blocks"),
	"diff": (print_record_diff, "list the record diff between two CAR files")
}

def print_help():
	print("USAGE: cartool COMMAND [args...]")
	print("")
	print("Available commands:")
	for cmdname, (cmdfn, helptext) in COMMANDS.items():
		fn_args = cmdfn.__code__.co_varnames[:cmdfn.__code__.co_argcount]
		args = [f"<{arg}>" for arg in fn_args]
		print(f"{cmdname} {' '.join(args)} : {helptext}")

def main():
	if len(sys.argv) < 2:
		print_help()
		return

	command, *args = sys.argv[1:]
	COMMANDS[command][0](*args)

if __name__ == "__main__":
	main()



from typing import Tuple, Optional, Any

from cbrrr import CID

from .node import MSTNode
from .node_store import NodeStore

# tuple helpers
def _tuple_replace_at(original: tuple, i: int, value: Any) -> tuple:
	return original[:i] + (value,) + original[i + 1:]

def _tuple_insert_at(original: tuple, i: int, value: Any) -> tuple:
	return original[:i] + (value,) + original[i:]

def _tuple_remove_at(original: tuple, i: int) -> tuple:
	return original[:i] + original[i + 1:]


class NodeWrangler:
	"""
	NodeWrangler is where core MST transformation ops are implemented, backed
	by a NodeStore

	The external APIs take a CID (the MST root) and return a CID (the new root),
	while storing any newly created nodes in the NodeStore.

	Neither method should ever fail - deleting a node that doesn't exist is a nop,
	and adding the same node twice with the same value is also a nop. Callers
	can detect these cases by seeing if the initial and final CIDs changed.
	"""
	ns: NodeStore

	def __init__(self, ns: NodeStore) -> None:
		self.ns = ns

	def put_record(self, root_cid: CID, key: str, val: CID) -> CID:
		root = self.ns.get_node(root_cid)
		if root.is_empty(): # special case for empty tree
			return self._put_here(root, key, val).cid
		return self._put_recursive(root, key, val, MSTNode.key_height(key), root.definitely_height()).cid

	def del_record(self, root_cid: CID, key: str) -> CID:
		root = self.ns.get_node(root_cid)

		# Note: the seemingly redundant outer .get().cid is required to transform
		# a None cid into the cid representing an empty node (we could maybe find a more elegant
		# way of doing this...)
		return self.ns.get_node(self._squash_top(
			self._delete_recursive(
				root, key, MSTNode.key_height(key), root.definitely_height()
			)
		)).cid



	def _put_here(self, node: MSTNode, key: str, val: CID) -> MSTNode:
		i = node.gte_index(key)

		# the key is already present!
		if i < len(node.keys) and node.keys[i] == key:
			if node.vals[i] == val:
				return node # we can return our old self if there is no change
			return self.ns.stored_node(MSTNode(
				keys=node.keys,
				vals=_tuple_replace_at(node.vals, i, val),
				subtrees=node.subtrees
			))
		
		return self.ns.stored_node(MSTNode(
			keys=_tuple_insert_at(node.keys, i, key),
			vals=_tuple_insert_at(node.vals, i, val),
			subtrees = node.subtrees[:i] + \
				self._split_on_key(node.subtrees[i], key) + \
				node.subtrees[i + 1:],
		))
	
	def _put_recursive(self, node: MSTNode, key: str, val: CID, key_height: int, tree_height: int) -> MSTNode:
		if key_height > tree_height: # we need to grow the tree
			return self.ns.stored_node(self._put_recursive(
				self.ns.stored_node(MSTNode(
					keys=(),
					vals=(),
					subtrees=(node.cid,)
				)),
				key, val, key_height, tree_height + 1
			))
		
		if key_height < tree_height: # we need to look below
			i = node.gte_index(key)
			return self.ns.stored_node(MSTNode(
				keys=node.keys,
				vals=node.vals,
				subtrees=_tuple_replace_at(
					node.subtrees, i,
					self._put_recursive(
						self.ns.get_node(node.subtrees[i]),
						key, val, key_height, tree_height - 1
					).cid
				)
			))
		
		# we can insert here
		assert(key_height == tree_height)
		return self._put_here(node, key, val)
	
	def _split_on_key(self, node_cid: Optional[CID], key: str) -> Tuple[Optional[CID], Optional[CID]]:
		if node_cid is None:
			return None, None
		node = self.ns.get_node(node_cid)
		i = node.gte_index(key)
		lsub, rsub = self._split_on_key(node.subtrees[i], key)
		return self.ns.stored_node(MSTNode(
			keys=node.keys[:i],
			vals=node.vals[:i],
			subtrees=node.subtrees[:i] + (lsub,)
		))._to_optional(), self.ns.stored_node(MSTNode(
			keys=node.keys[i:],
			vals=node.vals[i:],
			subtrees=(rsub,) + node.subtrees[i + 1:],
		))._to_optional()

	def _squash_top(self, node_cid: Optional[CID]) -> Optional[CID]:
		"""
		strip empty nodes from the top of the tree
		"""
		node = self.ns.get_node(node_cid)
		if node.keys:
			return node_cid
		if node.subtrees[0] is None:
			return node_cid
		return self._squash_top(node.subtrees[0])

	def _delete_recursive(self, node: MSTNode, key: str, key_height: int, tree_height: int) -> Optional[CID]:
		if key_height > tree_height: # the key cannot possibly be in this tree, no change needed
			return node._to_optional()
		
		i = node.gte_index(key)
		if key_height < tree_height: # the key must be deleted from a subtree
			if node.subtrees[i] is None:
				return node._to_optional() # the key cannot be in this subtree, no change needed
			return self.ns.stored_node(MSTNode(
				keys=node.keys,
				vals=node.vals,
				subtrees=_tuple_replace_at(
					node.subtrees,
					i,
					self._delete_recursive(self.ns.get_node(node.subtrees[i]), key, key_height, tree_height - 1)
				)
			))._to_optional()
		
		i = node.gte_index(key)
		if i == len(node.keys) or node.keys[i] != key:
			return node._to_optional() # key already not present
		
		assert(node.keys[i] == key) # sanity check (should always be true)

		return self.ns.stored_node(MSTNode(
			keys=_tuple_remove_at(node.keys, i),
			vals=_tuple_remove_at(node.vals, i),
			subtrees=node.subtrees[:i] + (
				self._merge(node.subtrees[i], node.subtrees[i + 1]),
			) + node.subtrees[i + 2:]
		))._to_optional()
	
	def _merge(self, left_cid: Optional[CID], right_cid: Optional[CID]) -> Optional[CID]:
		if left_cid is None:
			return right_cid # includes the case where left == right == None
		if right_cid is None:
			return left_cid
		left = self.ns.get_node(left_cid)
		right = self.ns.get_node(right_cid)
		return self.ns.stored_node(MSTNode(
			keys=left.keys + right.keys,
			vals=left.vals + right.vals,
			subtrees=left.subtrees[:-1] + (
				self._merge(
					left.subtrees[-1],
					right.subtrees[0]
				),
			 ) + right.subtrees[1:]
		))._to_optional()



import unittest
import random

from atmst.all import MemoryBlockStore, NodeStore, NodeWrangler, mst_diff, very_slow_mst_diff
from atmst.mst.node import MSTNode
from cbrrr import CID

def dump_mst(ns: NodeStore, cid: CID, lvl=0):
	node = ns.get_node(cid)
	print("  "*lvl + "-", node)
	for subtree in node.subtrees:
		if subtree:
			dump_mst(ns, subtree, lvl+1)

class MSTDiffTestCase(unittest.TestCase):
	def setUp(self):
		keys = []
		dummy_value = CID.cidv1_dag_cbor_sha256_32_from(b"value")
		i = 0
		for height in [0, 1, 0, 2, 0, 1, 0]: # if all these keys are added to a MST, it'll form a perfect binary tree.
			while True:
				key = f"k/{i:02d}"
				i += 1
				if MSTNode.key_height(key) == height:
					keys.append(key)
					break

		#print(keys)

		bs = MemoryBlockStore()
		self.ns = NodeStore(bs)
		wrangler = NodeWrangler(self.ns)

		# create all possible permutations of the full binary tree
		# the idea is that this'll cover most "interesting" trees up to a height of 3

		self.trees = []
		for i in range(2**len(keys)):
			root = self.ns.get_node(None).cid
			for j, k in enumerate(keys):
				if (i>>j)&1:
					root = wrangler.put_record(root, k, dummy_value)
			self.trees.append(root)
	
	def test_diff_all_pairs(self):
		for ai, a in enumerate(self.trees):
			for bi, b in enumerate(self.trees):
				#print(ai, bi)
				#print(self.ns.pretty(a))
				#print(self.ns.pretty(b))
				reference_created, reference_deleted = very_slow_mst_diff(self.ns, a, b)
				created, deleted = mst_diff(self.ns, a, b)
				self.assertEqual(created, reference_created)
				self.assertEqual(deleted, reference_deleted)
	
	def test_insertion_order_independent(self):
		wrangler = NodeWrangler(self.ns)

		keys = [str(x) for x in range(1000)]

		mst_a = MSTNode.empty_root().cid
		mst_b = MSTNode.empty_root().cid
		mst_c = MSTNode.empty_root().cid

		for k in keys:
			mst_a = wrangler.put_record(mst_a, k, CID.cidv1_dag_cbor_sha256_32_from(k.encode()))

		for k in keys[::-1]:
			mst_b = wrangler.put_record(mst_b, k, CID.cidv1_dag_cbor_sha256_32_from(k.encode()))

		random.shuffle(keys)
		for k in keys:
			mst_c = wrangler.put_record(mst_c, k, CID.cidv1_dag_cbor_sha256_32_from(k.encode()))

		#print()
		#dump_mst(self.ns, mst_a)

		#print()
		#dump_mst(self.ns, mst_b)

		self.assertEqual(mst_a, mst_b)
		self.assertEqual(mst_a, mst_c)

if __name__ == '__main__':
	unittest.main(module="tests.test_mst_diff")



from typing import Tuple, Set, Dict, Iterable, Optional
from enum import Enum
from dataclasses import dataclass
import json

from cbrrr import CID

from .node import MSTNode
from .node_store import NodeStore
from .node_walker import NodeWalker

class DeltaType(Enum):
	CREATED = 1
	UPDATED = 2
	DELETED = 3

@dataclass
class RecordDelta:
	delta_type: DeltaType
	path: str
	prior_value: Optional[CID]
	later_value: Optional[CID]

	def __repr__(self) -> str:
		prior = "NULL" if self.prior_value is None else self.prior_value.encode('base32')
		later = "NULL" if self.later_value is None else self.later_value.encode('base32')
		return f"{self.delta_type.name} {json.dumps(self.path)}: {prior} -> {later}"

def record_diff(ns: NodeStore, created: set[CID], deleted: set[CID]) -> Iterable[RecordDelta]:
	"""
	Given two sets of MST nodes (for example, the result of :meth:`mst_diff`), this
	returns an iterator of record changes.
	"""
	created_kv: Dict[str, CID] = dict(sum((list(zip(node.keys, node.vals)) for node in map(ns.get_node, created)), []))
	deleted_kv: Dict[str, CID] = dict(sum((list(zip(node.keys, node.vals)) for node in map(ns.get_node, deleted)), []))
	for created_key in created_kv.keys() - deleted_kv.keys():
		yield RecordDelta(
			delta_type=DeltaType.CREATED,
			path=created_key,
			prior_value=None,
			later_value=created_kv[created_key]
		)
	for updated_key in created_kv.keys() & deleted_kv.keys():
		v1 = deleted_kv[updated_key]
		v2 = created_kv[updated_key]
		if v1 != v2:
			yield RecordDelta(
				delta_type=DeltaType.UPDATED,
				path=updated_key,
				prior_value=v1,
				later_value=v2
			)
	for deleted_key in deleted_kv.keys() - created_kv.keys():
		yield RecordDelta(
			delta_type=DeltaType.DELETED,
			path=deleted_key,
			prior_value=deleted_kv[deleted_key],
			later_value=None
		)

def very_slow_mst_diff(ns: NodeStore, root_a: CID, root_b: CID):
	"""
	This should return the same result as :meth:`mst_diff`, but it gets there in a slow
	but much more obvious way (enumerating all nodes), so it's useful for testing.

	It's actually faster for smaller trees, but it chokes on trees with thousands of nodes (especially if the NodeStore is slow).
	"""
	a_nodes = set(NodeWalker(ns, root_a).iter_node_cids())
	b_nodes = set(NodeWalker(ns, root_b).iter_node_cids())
	return b_nodes - a_nodes, a_nodes - b_nodes

EMPTY_NODE_CID = MSTNode.empty_root().cid

def mst_diff(ns: NodeStore, root_a: CID, root_b: CID) -> Tuple[Set[CID], Set[CID]]: # created, deleted
	"""
	XXX: This implementation is not yet ready for prime-time!

	Given two MST root node CIDs, efficiently compute the difference between the two trees. The result is two sets, holding the created and deleted MST nodes respectively (referenced by CIDs).
	"""
	created = set() # MST nodes in b but not in a
	deleted = set() # MST nodes in a but not in b
	_mst_diff_recursive(created, deleted, NodeWalker(ns, root_a), NodeWalker(ns, root_b))
	middle = created & deleted # my algorithm has occasional false-positives
	#assert(not middle) # this fails
	#print("middle", len(middle))
	created -= middle
	deleted -= middle
	# special case: if one of the root nodes was empty
	if root_a == EMPTY_NODE_CID and root_b != EMPTY_NODE_CID:
		deleted.add(EMPTY_NODE_CID)
	if root_b == EMPTY_NODE_CID and root_a != EMPTY_NODE_CID:
		created.add(EMPTY_NODE_CID)
	return created, deleted

def _mst_diff_recursive(created: Set[CID], deleted: Set[CID], a: NodeWalker, b: NodeWalker): # created, deleted
	# the easiest of all cases
	if a.frame.node == b.frame.node:
		return # no difference
	
	# trivial
	if a.frame.node.is_empty():
		#mst_deleted.add(a.frame.node.cid) # this doesn't work because it might've been a null subtree node
		created |= set(b.iter_node_cids())
		return
	
	# likewise
	if b.frame.node.is_empty():
		#mst_created.add(b.frame.node.cid)
		deleted |= set(a.iter_node_cids())
		return
	
	# now we're onto the hard part

	"""
	theory: most trees that get compared will have lots of shared blocks (which we can skip over, due to identical CIDs)
	completely different trees will inevitably have to visit every node.

	general idea:
	1. if one cursor is "behind" the other, catch it up
	2. when we're matched up, skip over identical subtrees (and recursively diff non-identical subtrees)

	XXX: this seems to work nicely but I'm not sure if it's necessarily efficient for all tree layouts?
	"""

	# NB: these will end up as false-positives if one tree is a subtree of the other
	created.add(b.frame.node.cid)
	deleted.add(a.frame.node.cid)

	while True:
		while a.rpath != b.rpath: # we need a loop because they might "leapfrog" each other
			# "catch up" cursor a, if it's behind
			while a.rpath < b.rpath and not a.is_final:
				if a.subtree: # recurse down every subtree
					a.down()
					deleted.add(a.frame.node.cid)
				else:
					a.right_or_up()
			
			# catch up cursor b, likewise
			while b.rpath < a.rpath and not b.is_final:
				if b.subtree: # recurse down every subtree
					b.down()
					created.add(b.frame.node.cid)
				else:
					b.right_or_up()

		# the rpaths now match, but the subrees below us might not
		
		_mst_diff_recursive(created, deleted, a.subtree_walker(), b.subtree_walker())

		# check if we can still go right XXX: do we need to care about the case where one can, but the other can't?
		# To consider: maybe if I just step a, b will catch up automagically
		if a.rpath == a.stack[0].rpath and b.rpath == b.stack[0].rpath:
			break

		a.right_or_up()
		b.right_or_up()



from typing import Dict, List, Tuple, BinaryIO
import hashlib

from cbrrr import decode_dag_cbor, CID

from . import BlockStore

# should be equivalent to multiformats.varint.decode(), but not extremely slow for no reason.
def decode_varint(stream: BinaryIO):
	n = 0
	for shift in range(0, 63, 7):
		val = stream.read(1)
		if not val:
			raise ValueError("unexpected end of varint input")
		val = val[0]
		n |= (val & 0x7f) << shift
		if not val & 0x80:
			if shift and not val:
				raise ValueError("varint not minimally encoded")
			return n
		shift += 7
	raise ValueError("varint too long")

def encode_varint(n: int) -> bytes:
	if not 0 <= n < 2**63:
		raise ValueError("integer out of encodable varint range")
	res = []
	while n > 0x7f:
		res.append(0x80 | (n & 0x7f))
		n >>= 7
	res.append(n)
	return bytes(res)

class ReadOnlyCARBlockStore(BlockStore):
	"""
	This is a sliiiightly unclean abstraction because BlockStores are indexed
	by `bytes` rather than CID, but same idea. This is convenient for verifying
	proofs provided in CAR format, and for testing.
	"""

	car_root: CID
	block_offsets: Dict[bytes, Tuple[int, int]] # CID -> (offset, length)

	def __init__(self, file: BinaryIO, validate_hashes: bool=True) -> None:
		"""
		pre-scan over the whole file, recording the offsets of each block
		"""

		self.file = file
		self.validate_hashes = validate_hashes
		file.seek(0)

		# parse out CAR header
		header_len = decode_varint(file)
		header = file.read(header_len)
		if len(header) != header_len:
			raise EOFError("not enough CAR header bytes")
		header_obj = decode_dag_cbor(header)
		if header_obj.get("version") != 1:
			raise ValueError(f"unsupported CAR version ({header_obj.get('version')})")
		if len(header_obj["roots"]) != 1:
			raise ValueError(f"unsupported number of CAR roots ({len(header_obj['roots'])}, expected 1)")
		self.car_root = header_obj["roots"][0]

		# scan through the CAR to find block offsets
		self.block_offsets = {}
		while True:
			try:
				length = decode_varint(file)
			except ValueError:
				break # EOF
			start = file.tell()
			CID_LENGTH = 36  # XXX: this is a questionable assumption!!!
			cid = CID(file.read(CID_LENGTH))
			if not cid.is_cidv1_dag_cbor_sha256_32(): # I think this is enough to verify the assumption
				raise ValueError("unsupported CID type")
			self.block_offsets[bytes(cid)] = (start + CID_LENGTH, length - CID_LENGTH)
			file.seek(start + length)
	
	def put_block(self, key: bytes, value: bytes) -> None:
		raise NotImplementedError("ReadOnlyCARBlockStore does not support put()")
	
	def get_block(self, key: bytes) -> bytes:
		offset, length = self.block_offsets[key]
		self.file.seek(offset)
		value = self.file.read(length)
		if len(value) != length:
			raise EOFError()
		if self.validate_hashes:
			if key[:4] != CID.CIDV1_DAG_CBOR_SHA256_32_PFX:
				raise ValueError("unsupported CID type")
			digest = hashlib.sha256(value).digest()
			if digest != key[4:]:
				raise ValueError("bad CID hash!")
		return value
	
	def del_block(self, key: bytes) -> None:
		raise NotImplementedError("ReadOnlyCARBlockStore does not support delete()")



from typing import Optional, Dict
from functools import lru_cache

from cbrrr import CID
from lru import LRU

from ..blockstore import BlockStore
from ..util import indent
from .node import MSTNode

class NodeStore:
	"""
	NodeStore wraps a BlockStore to provide a more ergonomic interface
	for loading and storing MSTNodes
	"""
	bs: BlockStore
	cache: Dict[Optional[CID], MSTNode]

	def __init__(self, bs: BlockStore) -> None:
		self.bs = bs
		self.cache = LRU(1024)
	
	def get_node(self, cid: Optional[CID]) -> MSTNode:
		if cached := self.cache.get(cid): # look in our LRU cache first
			return cached
		"""
		if cid is None, returns an empty MST node
		"""
		if cid is None:
			return self.stored_node(MSTNode.empty_root())
		
		node_bytes = self.bs.get_block(bytes(cid))
		node = MSTNode.deserialise(node_bytes)

		# prime the cached_properties since we already know their values
		object.__setattr__(node, "serialised", node_bytes)
		object.__setattr__(node, "cid", cid)

		# prime the node cache
		self.cache[cid] = node

		return node
	
	def stored_node(self, node: MSTNode) -> MSTNode:
		self.cache[node.cid] = node # also put it in the LRU cache
		self.bs.put_block(bytes(node.cid), node.serialised)
		return node # this is convenient

	# MST pretty-printing
	# this should maybe not be implemented here
	def pretty(self, node_cid: Optional[CID]) -> str:
		if node_cid is None:
			return "<empty>"
		node = self.get_node(node_cid)
		res = f"MSTNode<cid={node.cid.encode('base32')}, maybe_height={node.maybe_height}>(\n{indent(self.pretty(node.subtrees[0]))},\n"
		for k, v, t in zip(node.keys, node.vals, node.subtrees[1:]):
			res += f"  {k!r} ({MSTNode.key_height(k)}) -> {v.encode('base32')},\n"
			res += indent(self.pretty(t)) + ",\n"
		res += ")"
		return res



from abc import ABC, abstractmethod
from typing import Optional, Dict
import sqlite3


class BlockStore(ABC):
	"""
	A block store is a k/v store where values are immutable once set. They can be deleted, though.
	In practice, k==hash(v), but this API doesn't care about that.

	I'm not using the "native" __getitem__, __setitem__, __del__ methods because
	the semantics of these methods differ subtly.

	if you call put() twice with the same args, the second call is a nop.
	if you call put() twice with the same key but different value, you get a ValueError

	get() offers no default return value, you get a KeyError if it doesn't exist.

	if you try to delete a key that doesn't exist, that's a nop.
	"""

	@abstractmethod
	def put_block(self, key: bytes, value: bytes) -> None:
		pass

	@abstractmethod
	def get_block(self, key: bytes) -> bytes:
		pass

	@abstractmethod
	def del_block(self, key: bytes) -> None:
		pass


class MemoryBlockStore(BlockStore):
	_state = Dict[bytes, bytes]

	def __init__(self, state: Optional[Dict[bytes, bytes]]=None) -> None:
		"""
		NB: if a state dict is passed, it'll get mutated in-place
		"""
		self._state = dict() if state is None else state
	
	def put_block(self, key: bytes, value: bytes) -> None:
		existing_value = self._state.get(key)
		if existing_value:
			if existing_value == value:
				return  # the value matches, there's nothing to do
			raise ValueError("block values are immutable")
		self._state[key] = value
	
	def get_block(self, key: bytes) -> bytes:
		value = self._state.get(key)
		if value is None:
			raise KeyError("no block matches this key")
		return value
	
	def del_block(self, key: bytes) -> None:
		if key in self._state:
			del self._state[key]


class SqliteBlockStore(BlockStore):
	"""
	NB: Caller is responsible for calling commit(), etc.
	TODO: consider allowing a custom table name?
	"""
	def __init__(self, con: sqlite3.Connection, table: str="mst_blocks") -> None:
		self.table = table
		self._cur = con.cursor()
		self._cur.execute(f"""
			CREATE TABLE IF NOT EXISTS {self.table} (
				block_key BLOB PRIMARY KEY,
				block_val BLOB NOT NULL
			) WITHOUT ROWID;
		""")
	
	def put_block(self, key: bytes, value: bytes) -> None:
		# XXX: this will fail silently if the key already exists but with a different value
		# (that should never happen but it'd be nice to have guard rails)
		self._cur.execute(f"INSERT OR IGNORE INTO {self.table} (block_key, block_val) VALUES (?, ?)", (key, value))
	
	def get_block(self, key: bytes) -> bytes:
		row = self._cur.execute(f"SELECT block_val FROM {self.table} WHERE block_key=?", (key,)).fetchone()
		if row is None:
			raise KeyError("no block matches this key")
		return row[0]
	
	def del_block(self, key: bytes) -> None:
		self._cur.execute(f"DELETE FROM {self.table} WHERE block_key=?", (key,))


class OverlayBlockStore(BlockStore):
	"""
	reads come from "upper", then "lower" if they don't exist in upper.
	writes/deletes go only to "upper".
	"""

	def __init__(self, upper: BlockStore, lower: BlockStore) -> None:
		self.upper = upper
		self.lower = lower
	
	def put_block(self, key: bytes, value: bytes) -> None:
		self.upper.put_block(key, value)
	
	def get_block(self, key: bytes) -> bytes:
		try:
			return self.upper.get_block(key)
		except KeyError:
			return self.lower.get_block(key)
	
	def del_block(self, key: bytes) -> None:
		self.upper.del_block(key)


"""
if __name__ == "__main__":
	import os

	bs = MemoryBlockStore()
	bs.put_block(b"hello", b"world")

	bs.put_block(b"hello", b"world") # putting twice is a nop

	try:
		bs.put_block(b"hello", b"foobar")
		assert(False) # should be unreachable
	except ValueError:
		pass

	print("hello ->", bs.get_block(b"hello"))

	bs.del_block(b"nothing") # nop

	bs.del_block(b"hello")

	try:
		bs.get_block(b"hello")
		assert(False) # should be unreachable
	except KeyError:
		pass

	TEST_DB = "test.db"

	with sqlite3.connect(TEST_DB) as db:
		bs = SqliteBlockStore(db)
		bs.put_block(b"hello", b"sqlite world")

	with sqlite3.connect(TEST_DB) as db:
		bs = SqliteBlockStore(db)
		print("hello ->", bs.get_block(b"hello"))
		bs.del_block(b"hello")
	
	try:
		with sqlite3.connect(TEST_DB) as db:
			bs = SqliteBlockStore(db)
			print("hello ->", bs.get_block(b"hello"))
		assert(False) # should be unreachable
	except KeyError:
		pass

	os.remove(TEST_DB) # clean up
"""



import hashlib
import operator
from functools import cached_property
from more_itertools import ilen
from itertools import takewhile
from dataclasses import dataclass

from typing import TYPE_CHECKING, Tuple, Optional
if TYPE_CHECKING: # Self doesn't exist <3.11
	from typing import Self

from cbrrr import encode_dag_cbor, decode_dag_cbor, CID


@dataclass(frozen=True) # frozen == immutable == win
class MSTNode:
	"""
	k/v pairs are interleaved between subtrees like so: ::

		keys:          (0,    1,    2,    3)
		vals:          (0,    1,    2,    3)
		subtrees:   (0,    1,    2,    3,    4)

	If a method is implemented in this class, it's because it's a function/property
	of a single node, as opposed to a whole tree
	"""
	keys: Tuple[str] # collection/rkey
	vals: Tuple[CID] # record CIDs
	subtrees: Tuple[Optional[CID]] # a None value represents an empty subtree


	# NB: __init__ is auto-generated by dataclass decorator

	# these checks could be elided for "trusted" trees
	def __post_init__(self) -> None:
		# TODO: maybe check that they're tuples here?
		# implicitly, the length of self.subtrees must be at least 1
		if len(self.subtrees) != len(self.keys) + 1:
			raise ValueError("Invalid subtree count")
		if len(self.keys) != len(self.vals):
			raise ValueError("Mismatched keys/vals lengths")
		key_heights = [MSTNode.key_height(k) for k in self.keys]
		if any(k!=key_heights[0] for k in key_heights):
			raise ValueError("Inconsistent key heights")

	@classmethod
	def empty_root(cls) -> "Self":
		return cls(
			subtrees=(None,),
			keys=(),
			vals=()
		)

	# this should maybe not be implemented here?
	@staticmethod
	def key_height(key: str) -> int:
		digest = int.from_bytes(hashlib.sha256(key.encode()).digest(), "big")
		leading_zeroes = 256 - digest.bit_length()
		return leading_zeroes // 2

	# since we're immutable, this can be cached
	@cached_property
	def cid(self) -> CID:
		return CID.cidv1_dag_cbor_sha256_32_from(self.serialised)

	# likewise
	@cached_property
	def serialised(self) -> bytes:
		e = []
		prev_key = b""
		for subtree, key_str, value in zip(self.subtrees[1:], self.keys, self.vals):
			key_bytes = key_str.encode()
			shared_prefix_len = ilen(takewhile(bool, map(operator.eq, prev_key, key_bytes))) # I love functional programming
			e.append({
				"k": key_bytes[shared_prefix_len:],
				"p": shared_prefix_len,
				"t": subtree,
				"v": value,
			})
			prev_key = key_bytes
		return encode_dag_cbor({
			"e": e,
			"l": self.subtrees[0]
		})

	@classmethod
	def deserialise(cls, data: bytes) -> "Self":
		cbor = decode_dag_cbor(data)
		if len(cbor) != 2: # e, l
			raise ValueError("malformed MST node")
		subtrees = [cbor["l"]]
		keys = []
		vals = []
		prev_key = b""
		for e in cbor["e"]: # TODO: make extra sure that these checks are watertight wrt non-canonical representations
			if len(e) != 4: # k, p, t, v
				raise ValueError("malformed MST node")
			prefix_len: int = e["p"]
			suffix: bytes = e["k"]
			if prefix_len > len(prev_key):
				raise ValueError("invalid MST key prefix len")
			if prev_key[prefix_len:prefix_len+1] == suffix[:1]:
				raise ValueError("non-optimal MST key prefix len")
			this_key = prev_key[:prefix_len] + suffix
			if this_key <= prev_key:
				raise ValueError("invalid MST key sort order")
			keys.append(this_key.decode())
			vals.append(e["v"])
			subtrees.append(e["t"])
			prev_key = this_key

		return cls(
			subtrees=tuple(subtrees),
			keys=tuple(keys),
			vals=tuple(vals)
		)

	def is_empty(self) -> bool:
		return self.subtrees == (None,)

	def _to_optional(self) -> Optional[CID]:
		"""
		returns None if the node is empty
		"""
		if self.is_empty():
			return None
		return self.cid


	@cached_property
	def maybe_height(self) -> Optional[int]:
		# if there are keys at this level, query one directly
		if self.keys:
			return self.key_height(self.keys[0])

		# we're an empty tree
		if self.subtrees[0] is None:
			return 0

		# this should only happen for non-root nodes with no keys (aka an empty intermediate node)
		return None
		# NOTE: a Node class cannot see what's below it. You'll need to track
		# state externally (like NodeWrangler does) if you want to find out

	def definitely_height(self) -> int:
		if self.maybe_height is None:
			raise ValueError("indeterminate node height")
		return self.maybe_height

	def gte_index(self, key: str) -> int:
		"""
		find the index of the first key greater than or equal to the specified key
		if all keys are smaller, it returns len(keys)
		"""
		i = 0 # this loop could be a binary search but not worth it for small fanouts
		while i < len(self.keys) and key > self.keys[i]:
			i += 1
		return i


"""
if __name__ == "__main__":
	from .blockstore import MemoryBlockStore, OverlayBlockStore
	from .blockstore.car_reader import ReadOnlyCARBlockStore

	if 0:
		import sys
		sys.setrecursionlimit(999999999)
		f = open("/home/david/programming/python/bskyclient/retr0id.car", "rb")
		bs = OverlayBlockStore(MemoryBlockStore(), ReadOnlyCARBlockStore(f))
		commit_obj = dag_cbor.decode(bs.get_block(bytes(bs.lower.car_roots[0])))
		mst_root: CID = commit_obj["data"]
		ns = NodeStore(bs)
		wrangler = NodeWrangler(ns)
		#print(wrangler)
		#enumerate_mst(ns, mst_root)
		enumerate_mst_range(ns, mst_root, "app.bsky.feed.generator/", "app.bsky.feed.generator/\xff")

		root2 = wrangler.del_record(mst_root, "app.bsky.feed.generator/alttext")
		root2 = wrangler.del_record(root2, "app.bsky.feed.like/3kas3fyvkti22")
		root2 = wrangler.put_record(root2, "app.bsky.feed.like/3kc3brpic2z2p", hash_to_cid(b"blah"))

		c, d = mst_diff(ns, mst_root, root2)
		print("CREATED:")
		for x in c:
			print("created", x.encode("base32"))
		print("DELETED:")
		for x in d:
			print("deleted", x.encode("base32"))

		for op in record_diff(ns, c, d):
			print(op)
		
		e, f = very_slow_mst_diff(ns, mst_root, root2)
		assert(e == c)
		assert(f == d)
	else:
		bs = MemoryBlockStore()
		ns = NodeStore(bs)
		wrangler = NodeWrangler(ns)
		root = ns.get_node(None).cid
		print(ns.pretty(root))
		root = wrangler.put_record(root, "hello", hash_to_cid(b"blah"))
		print(ns.pretty(root))
		root = wrangler.put_record(root, "foo", hash_to_cid(b"bar"))
		print(ns.pretty(root))
		root_a = root
		root = wrangler.put_record(root, "bar", hash_to_cid(b"bat"))
		root = wrangler.put_record(root, "xyzz", hash_to_cid(b"bat"))
		root = wrangler.del_record(root, "foo")
		print("=============")
		print(ns.pretty(root_a))
		print("=============")
		print(ns.pretty(root))
		#exit()
		print("=============")
		enumerate_mst(ns, root)
		c, d = mst_diff(ns, root_a, root)
		print("CREATED:")
		for x in c:
			print("created", x.encode("base32"))
		print("DELETED:")
		for x in d:
			print("deleted", x.encode("base32"))
		
		e, f = very_slow_mst_diff(ns, root_a, root)
		assert(e == c)
		assert(f == d)
		
		exit()
		root = wrangler.delete(root, "foo")
		root = wrangler.delete(root, "hello")
		print(ns.pretty(root))
		root = wrangler.delete(root, "bar")
		print(ns.pretty(root))
		root = wrangler.delete(root, "bar")
		print(ns.pretty(root))
"""



"""
This module collects "all" the commonly needed imports into a single import.
It might be convenient. I might get rid of it.
"""

from .blockstore import BlockStore, MemoryBlockStore
from .blockstore.car_file import ReadOnlyCARBlockStore
from .mst.node_walker import NodeWalker
from .mst.node_store import NodeStore
from .mst.node_wrangler import NodeWrangler
from .mst.diff import mst_diff, very_slow_mst_diff, record_diff

__all__ = [
	"BlockStore", "MemoryBlockStore", "ReadOnlyCARBlockStore",
	"NodeWalker", "NodeStore", "NodeWrangler",
	"mst_diff", "very_slow_mst_diff", "record_diff",
]



from dataclasses import dataclass
from typing import TYPE_CHECKING, Tuple, Optional, List, Iterable
if TYPE_CHECKING: # Self doesn't exist <3.11
	from typing import Self

from cbrrr import CID

from .node import MSTNode
from .node_store import NodeStore

class NodeWalker:
	"""
	NodeWalker makes implementing tree diffing and other MST query ops more
	convenient (but it does not, itself, implement them).

	A NodeWalker starts off at the root of a tree, and can walk along or recurse
	down into subtrees.

	Walking "off the end" of a subtree brings you back up to its next non-empty parent.

	Recall MSTNode layout: ::

		keys:  (lpath)  (0,    1,    2,    3)  (rpath)
		vals:           (0,    1,    2,    3)
		subtrees:    (0,    1,    2,    3,    4)

	"""
	PATH_MIN = "" # string that compares less than all legal path strings
	PATH_MAX = "\xff" # string that compares greater than all legal path strings

	@dataclass
	class StackFrame:
		node: MSTNode # could store CIDs only to save memory, in theory, but not much point
		lpath: str
		rpath: str
		idx: int

	ns: NodeStore
	stack: List[StackFrame]
	root_height: int
	
	def __init__(self,
		ns: NodeStore,
		root_cid: Optional[CID],
		lpath: str=PATH_MIN,
		rpath: str=PATH_MAX,
		trusted: bool=False,
		root_height: Optional[int]=None
	) -> None:
		self.ns = ns
		self.trusted = trusted
		node = MSTNode.empty_root() if root_cid is None else self.ns.get_node(root_cid)
		self.root_height = node.maybe_height if root_height is None else root_height
		if self.root_height is None:
			raise ValueError("indeterminate node height - pass it in if you know it")
		self.stack = [self.StackFrame(
			node=node,
			lpath=lpath,
			rpath=rpath,
			idx=0
		)]
	
	def subtree_walker(self) -> "Self":
		return NodeWalker(
			self.ns,
			self.subtree,
			self.lpath,
			self.rpath,
			self.trusted,
			root_height=self.height - 1
		)
	
	@property
	def frame(self) -> StackFrame:
		return self.stack[-1]

	@property
	def height(self) -> int:
		return self.root_height - (len(self.stack) - 1)

	@property
	def lpath(self) -> str:
		return self.frame.lpath if self.frame.idx == 0 else self.frame.node.keys[self.frame.idx - 1]
	
	@property
	def lval(self) -> Optional[CID]:
		return None if self.frame.idx == 0 else self.frame.node.vals[self.frame.idx - 1]

	@property
	def subtree(self) -> Optional[CID]:
		return self.frame.node.subtrees[self.frame.idx]
	
	@property
	def rpath(self) -> str:
		return self.frame.rpath if self.frame.idx == len(self.frame.node.keys) else self.frame.node.keys[self.frame.idx]
	
	@property
	def rval(self) -> Optional[CID]:
		return None if self.frame.idx == len(self.frame.node.vals) else self.frame.node.vals[self.frame.idx]

	@property
	def is_final(self) -> bool:
		# is (not self.stack) really necesasry here? is that a reachable state?
		return (not self.stack) or (self.subtree is None and self.rpath == self.stack[0].rpath)

	@property
	def can_go_right(self) -> bool:
		return (self.frame.idx + 1) < len(self.frame.node.subtrees)

	def right_or_up(self) -> None:
		if not self.can_go_right:
			# we reached the end of this node, go up a level
			self.stack.pop() # TODO: check before pop - make empty-stack an unreachable state
			if not self.stack:
				raise StopIteration # you probably want to check .final instead of hitting this
			return self.right_or_up() # we need to recurse, to skip over empty intermediates on the way back up
		self.frame.idx += 1

	def right(self) -> None:
		if not self.can_go_right:
			raise ValueError("cursor is already at rightmost position in node")
		self.frame.idx += 1

	def down(self) -> None:
		subtree = self.frame.node.subtrees[self.frame.idx]
		if subtree is None:
			raise ValueError("oi, you can't recurse here mate (subtree is None)")

		subtree_node = self.ns.get_node(subtree)

		if not self.trusted: # if we "trust" the source we can elide this check
			# the "None" case occurs for empty intermediate nodes
			if subtree_node.maybe_height is not None and subtree_node.maybe_height != self.height - 1:
				raise ValueError(f"inconsistent subtree height ({subtree_node.maybe_height}, expected {self.height - 1})")

		self.stack.append(self.StackFrame(
			node=subtree_node,
			lpath=self.lpath,
			rpath=self.rpath,
			idx=0
		))
	
	# everything above here is core tree walking logic
	# everything below here is helper functions

	def next_kv(self) -> Tuple[str, CID]:
		while self.subtree: # recurse down every subtree
			self.down()
		self.right_or_up()
		return self.lpath, self.lval # the kv pair we just jumped over

	# iterate over every k/v pair in key-sorted order
	# NB: should really be p/v standing for path/value
	def iter_kv(self) -> Iterable[Tuple[str, CID]]:
		while not self.is_final:
			yield self.next_kv()
	
	# get all mst nodes down and to the right of the current position
	def iter_nodes(self) -> Iterable[MSTNode]:
		yield self.frame.node
		while not self.is_final:
			while self.subtree: # recurse down every subtree
				self.down()
				yield self.frame.node
			self.right_or_up()

	def iter_node_cids(self) -> Iterable[CID]:
		for node in self.iter_nodes():
			yield node.cid

	# start inclusive
	def iter_kv_range(self, start: str, end: str, end_inclusive: bool=False) -> Iterable[Tuple[str, CID]]:
		while True:
			while self.rpath < start:
				self.right_or_up()
			if not self.subtree:
				break
			self.down()

		for k, v, in self.iter_kv():
			if k > end or (not end_inclusive and k == end):
				break
			yield k, v

	# TODO: we need to make this early-exit so that it can work with concise deletion proofs, maybe
	# (early exit based on key height - might need significant rewrite)
	def find_rpath(self, rpath: str) -> Optional[CID]:
		rpath_height = MSTNode.key_height(rpath)
		while True:
			# if the rpath we're looking for is higher than the current cursor,
			# we're never going to find it (i.e. we early-exit)
			if rpath_height > self.height:
				return None
			while self.rpath < rpath: # either look for the rpath, or the right point to go down
				if not self.can_go_right:
					return None
				self.right()
			if self.rpath == rpath:
				return self.rval # found it!
			if not self.subtree:
				return None # need to go down, but we can't
			self.down()
